"""Remove the user's remote storage contents via CLI."""
import argparse

import inductiva
from inductiva.utils.input_functions import user_confirmation_prompt
from ...localization import translator as __


def remove(args):
    """Remove user's remote storage contents."""
    task_ids = args.id
    confirm = args.confirm

    task_ids = set(task_ids)
    if not confirm:
        confirm = user_confirmation_prompt(
            task_ids, __("storage-prompt-remove-all"),
            __("storage-prompt-remove-big", len(task_ids)),
            __("storage-prompt-remove-small"), False)

    if confirm:
        for task_id in task_ids:
            inductiva.tasks.Task(task_id).remove_remote_files()

    return 0


def register(parser):
    subparser = parser.add_parser("remove",
                                  aliases=["rm"],
                                  help="Remove remote storage entries.",
                                  formatter_class=argparse.RawTextHelpFormatter)
    subparser.description = (
        "The `inductiva storage remove` command deletes specified data"
        " from the platform.\n"
        "It targets a specific task files for removal. Use with caution as "
        "this action is irreversible.\n\n")
    subparser.add_argument("id",
                           type=str,
                           nargs="+",
                           help="Id(s) of the task(s) to remove from remote "
                           "storage.")

    subparser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        dest="confirm",
        default=False,
        help="Sets any confirmation values to \"yes\" "
        "automatically. Users will not be asked for "
        "confirmation to remove path(s) from remote storage.")

    subparser.set_defaults(func=remove)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class Project(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "num_tasks",
            "name",
            "created_at",
            "id",
        }
        
        class properties:
            
            
            class name(
                schemas.StrSchema
            ):
                pass
            id = schemas.UUIDSchema
            created_at = schemas.DateTimeSchema
            num_tasks = schemas.IntSchema
            
            
            class task_status_overview(
                schemas.DictSchema
            ):
            
            
                class MetaOapg:
                    additional_properties = schemas.IntSchema
                
                def __getitem__(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                    # dict_instance[name] accessor
                    return super().__getitem__(name)
                
                def get_item_oapg(self, name: typing.Union[str, ]) -> MetaOapg.additional_properties:
                    return super().get_item_oapg(name)
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[MetaOapg.additional_properties, decimal.Decimal, int, ],
                ) -> 'task_status_overview':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            __annotations__ = {
                "name": name,
                "id": id,
                "created_at": created_at,
                "num_tasks": num_tasks,
                "task_status_overview": task_status_overview,
            }
    
    num_tasks: MetaOapg.properties.num_tasks
    name: MetaOapg.properties.name
    created_at: MetaOapg.properties.created_at
    id: MetaOapg.properties.id
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["created_at"]) -> MetaOapg.properties.created_at: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["num_tasks"]) -> MetaOapg.properties.num_tasks: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["task_status_overview"]) -> MetaOapg.properties.task_status_overview: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["name", "id", "created_at", "num_tasks", "task_status_overview", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["name"]) -> MetaOapg.properties.name: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["created_at"]) -> MetaOapg.properties.created_at: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["num_tasks"]) -> MetaOapg.properties.num_tasks: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["task_status_overview"]) -> typing.Union[MetaOapg.properties.task_status_overview, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["name", "id", "created_at", "num_tasks", "task_status_overview", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        num_tasks: typing.Union[MetaOapg.properties.num_tasks, decimal.Decimal, int, ],
        name: typing.Union[MetaOapg.properties.name, str, ],
        created_at: typing.Union[MetaOapg.properties.created_at, str, datetime, ],
        id: typing.Union[MetaOapg.properties.id, str, uuid.UUID, ],
        task_status_overview: typing.Union[MetaOapg.properties.task_status_overview, dict, frozendict.frozendict, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'Project':
        return super().__new__(
            cls,
            *_args,
            num_tasks=num_tasks,
            name=name,
            created_at=created_at,
            id=id,
            task_status_overview=task_status_overview,
            _configuration=_configuration,
            **kwargs,
        )



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskStatusCode(
    schemas.EnumBase,
    schemas.StrSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Possible task status codes.
    """
    
    @schemas.classproperty
    def PENDINGINPUT(cls):
        return cls("pending-input")
    
    @schemas.classproperty
    def SUBMITTED(cls):
        return cls("submitted")
    
    @schemas.classproperty
    def STARTED(cls):
        return cls("started")
    
    @schemas.classproperty
    def SUCCESS(cls):
        return cls("success")
    
    @schemas.classproperty
    def FAILED(cls):
        return cls("failed")
    
    @schemas.classproperty
    def PENDINGKILL(cls):
        return cls("pending-kill")
    
    @schemas.classproperty
    def KILLED(cls):
        return cls("killed")
    
    @schemas.classproperty
    def SPOTINSTANCEPREEMPTED(cls):
        return cls("spot-instance-preempted")
    
    @schemas.classproperty
    def EXECUTERTERMINATED(cls):
        return cls("executer-terminated")
    
    @schemas.classproperty
    def EXECUTERTERMINATEDBYUSER(cls):
        return cls("executer-terminated-by-user")
    
    @schemas.classproperty
    def EXECUTERTERMINATEDTTLEXCEEDED(cls):
        return cls("executer-terminated-ttl-exceeded")
    
    @schemas.classproperty
    def EXECUTERFAILED(cls):
        return cls("executer-failed")
    
    @schemas.classproperty
    def ZOMBIE(cls):
        return cls("zombie")
    
    @schemas.classproperty
    def COMPUTATIONSTARTED(cls):
        return cls("computation-started")
    
    @schemas.classproperty
    def COMPUTATIONENDED(cls):
        return cls("computation-ended")
    
    @schemas.classproperty
    def TTLEXCEEDED(cls):
        return cls("ttl-exceeded")



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class DynamicDiskResizeConfig(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "free_space_threshold_gb",
            "size_increment_gb",
            "max_disk_size_gb",
        }
        
        class properties:
            free_space_threshold_gb = schemas.IntSchema
            size_increment_gb = schemas.IntSchema
            max_disk_size_gb = schemas.IntSchema
            __annotations__ = {
                "free_space_threshold_gb": free_space_threshold_gb,
                "size_increment_gb": size_increment_gb,
                "max_disk_size_gb": max_disk_size_gb,
            }
    
    free_space_threshold_gb: MetaOapg.properties.free_space_threshold_gb
    size_increment_gb: MetaOapg.properties.size_increment_gb
    max_disk_size_gb: MetaOapg.properties.max_disk_size_gb
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["free_space_threshold_gb"]) -> MetaOapg.properties.free_space_threshold_gb: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["size_increment_gb"]) -> MetaOapg.properties.size_increment_gb: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["max_disk_size_gb"]) -> MetaOapg.properties.max_disk_size_gb: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["free_space_threshold_gb", "size_increment_gb", "max_disk_size_gb", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["free_space_threshold_gb"]) -> MetaOapg.properties.free_space_threshold_gb: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["size_increment_gb"]) -> MetaOapg.properties.size_increment_gb: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["max_disk_size_gb"]) -> MetaOapg.properties.max_disk_size_gb: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["free_space_threshold_gb", "size_increment_gb", "max_disk_size_gb", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        free_space_threshold_gb: typing.Union[MetaOapg.properties.free_space_threshold_gb, decimal.Decimal, int, ],
        size_increment_gb: typing.Union[MetaOapg.properties.size_increment_gb, decimal.Decimal, int, ],
        max_disk_size_gb: typing.Union[MetaOapg.properties.max_disk_size_gb, decimal.Decimal, int, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'DynamicDiskResizeConfig':
        return super().__new__(
            cls,
            *_args,
            free_space_threshold_gb=free_space_threshold_gb,
            size_increment_gb=size_increment_gb,
            max_disk_size_gb=max_disk_size_gb,
            _configuration=_configuration,
            **kwargs,
        )



"""Functions to handle input from the user."""

from inductiva import constants
from ..localization import translator as __


def user_confirmation_prompt(items: list, all_msg: str, unlisted_msg: str,
                             listed_msg: str, is_all: bool) -> bool:
    """Prompt the user for confirmation to proceed with an action.

    Args:
        is_all (bool): Whether the action will affect all things.
        items (list): List of things that will be affected by
            the action.
        all_msg (str): Custom message when we select all things.
            Example: with all_things and all_msg="kill all tasks"
            we will print: "You are about to kill all tasks."
        unlisted_msg (str): Custom message when the list of
            things is too big.
            Example: with items with 100 elements and
            unlisted_msg="kill 100 tasks" we will print:
            "You are about to kill 100 tasks."
        listed_msg (str): Custom message when the list of things
            is small and we are about to list them all.
            Example: with items with 3 elements and
            listed_msg="kill the following tasks" we will print:
            "You are about to kill the following tasks:"
    Returns:
        bool: Whether the user has confirmed the action.
    """
    if is_all:
        print(all_msg)
    else:
        if len(items) > constants.MAX_CONFIRMATION_LINES:
            print(unlisted_msg)
        else:
            print(listed_msg)
            for thing in items:
                print(f"  - {thing}")
    prompt = input(__("user-prompt-confirmation"))
    confirm = prompt.lower() in ["y", "ye", "yes"]
    return confirm



"""OpenFOAM module of the API for fluid dynamics."""
from typing import Optional

from inductiva import types, tasks, simulators

AVAILABLE_OPENFOAM_DISTRIBUTIONS = ["foundation", "esi"]


@simulators.simulator.mpi_enabled
class OpenFOAM(simulators.Simulator):
    """Class to invoke a generic OpenFOAM simulation on the API.

    Users can choose between the ESI or the Foundation version
    by selecting the version on the initiliasation. Be aware, that
    some input files may only work for a specific version.
    """

    def __init__(self,
                 /,
                 distribution: str = "foundation",
                 version: Optional[str] = None,
                 use_dev: bool = False):
        """Initialize the OpenFOAM simulator.

        Args:
            distribution (str): The distribution of OpenFOAM to use. Available
                distributions are "foundation" and "esi". Default is
                "foundation".
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        if distribution not in AVAILABLE_OPENFOAM_DISTRIBUTIONS:
            raise ValueError(
                f"Distribution '{distribution}' of OpenFOAM is not supported. "
                f"Available distributions are: "
                f"{AVAILABLE_OPENFOAM_DISTRIBUTIONS}")

        self._distribution = distribution
        super().__init__(version=version, use_dev=use_dev)
        self.api_method_name = f"fvm.openfoam_{distribution}.run_simulation"

    @property
    def name(self):
        """Get the name of the simulator."""
        return "OpenFOAM-" + self._distribution

    def run(self,
            input_dir: str,
            commands: types.Commands,
            n_vcpus: Optional[int] = None,
            use_hwthread: bool = True,
            on: Optional[types.ComputationalResources] = None,
            storage_dir: Optional[str] = "",
            extra_metadata: Optional[dict] = None,
            resubmit_on_preemption: bool = False,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            commands: List of commands to run using the OpenFOAM simulator.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
            (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
            number of hardware threads on the node, and use that as the
            number of slots available.
            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiates with
                `spot=True`.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           n_vcpus=n_vcpus,
                           use_hwthread=use_hwthread,
                           extra_metadata=extra_metadata,
                           resubmit_on_preemption=resubmit_on_preemption,
                           **kwargs)



"""
Methods to interact with Benchmarks.
"""
from collections import defaultdict
import datetime
import logging
import time
import json

from typing import Any, Callable, Dict, List, Type, Union
from tqdm import tqdm

import inductiva
from inductiva.client import models
from inductiva.resources.machines_base import BaseMachineGroup
from inductiva.templating.manager import TemplateManager
from inductiva.resources.machines import MachineGroup
from inductiva.simulators.simulator import Simulator
from inductiva.projects.project import Project
from inductiva.resources import machine_groups
from inductiva.tasks.task import Task
from inductiva import users

from ..localization import translator as __

QUOTAS_EXCEEDED_SLEEP_TIME = 60


def _analyze_task(task: Task, report: Dict[str, Any], skipped_tasks: List[Task],
                  metadata_tasks: Dict[str, Any]):
    """Analyzes a task.

    This method analyzes a task and adds the information to the report. If the
    task is not successful, it is added to the skipped_tasks list.

    Args:
        task (Task): Task to analyze.
        report (Dict[str, Any]): Report to add the information to.
        skipped_tasks (List[Task]): List of tasks that were skipped.
        metadata_tasks (Dict[str, Any]): Dictionary with the metadata of the
            tasks.
    Returns:
        This method does not return anything. It modifies some parameters
        instead of returning multiple values.
    """
    task_status = task.get_status()
    if task_status != models.TaskStatusCode.SUCCESS:
        skipped_tasks.append(task)
        return
    task_info = task.info
    a, b, *_ = task_info.executer.vm_name.split("-")
    mg_name = a + "-" + b
    resource = machine_groups.get_by_name(mg_name)
    vm_type = task_info.executer.vm_type

    if vm_type not in report["resources"]:
        report["resources"][vm_type] = {
            "cost_per_hour": resource.estimate_cloud_cost(verbose=False),
            "type": resource.__class__.__name__,
            "provider": task_info.executer.host_type,
            "machine_count": resource.num_machines,
            "tasks": []
        }
    report["resources"][vm_type]["tasks"].append({
        "task_id": task_info.task_id,
        "info": task_info.to_dict(),
        "metadata": metadata_tasks[task_info.task_id]
    })


def analyze(name: str, metadata_path: str, ouput_path: str = None) -> Dict:
    """Analyzes a benchmark.
    This method creates a dictionary with all the relevant information about
    the benchmark.
    Args:
        name (str): Name of the benchmark.
        metadata_path (str): Path to the metadata file.
        ouput_path (str): Path to the output file.
            If None it will not create an output file.
    Returns:
        Dict: Dictionary with the information about the benchmark.
    """
    logging.info("Starting analysis of benchmark %s", name)
    project = Project(name)

    tasks = project.get_tasks()

    report = {"name": name, "resources": {}}

    #reads the metadata file and creates a dictionary with the task_id as
    #key and the metadata as value
    metadata_tasks = {}
    with open(metadata_path, "r", encoding="utf-8") as metadata_file:
        for line in metadata_file:
            temp_task = json.loads(line)
            metadata_tasks[temp_task["task_id"]] = temp_task
    skipped_tasks = []
    for task in tqdm(tasks, desc="Analyzing tasks"):
        _analyze_task(task, report, skipped_tasks, metadata_tasks)

    for task in skipped_tasks:
        logging.info("Task %s was skipped due to status %s", task.info.task_id,
                     task.info.status)
    if ouput_path:
        logging.info("Writing report to %s/%s.json", ouput_path, name)
        with open(f"{ouput_path}/{name}.json", "w",
                  encoding="utf-8") as output_file:
            output_file.write(json.dumps(report, indent=4))

    return report


def _tasks_by_vm_type(project: Project) -> Dict[str, List[Task]]:
    """Returns a dictionary with the resources and the tasks ran on them.
    Args:
        project (Project): Project to get the resources and tasks from.
    Returns:
        Dict[str, List[Task]]: Dictionary with the resources as keys and the
            tasks ran on them as values.
        Example:.
        { "n1-standard-4": [task1, task2], "n1-standard-8": [task3]}
    """
    tasks = project.get_tasks()
    resources = defaultdict(list)
    for task in tasks:
        task_info = task.get_info()
        if task_info.executer is None:
            raise RuntimeError(__("tasks_not_executed_yet"))

        resource = task_info.executer.vm_type
        resources[resource].append(task)
    return resources


def _compute_tasks_to_run(machines_to_run: List[str],
                          already_ran: Dict[str, List[Task]],
                          intended_replicas: int) -> Dict[str, int]:
    """Computes the number of tasks that need to be run for each resource.
    Args:
        machines_to_run (List[str]): List of machines to run the simulation on.
        already_ran (Dict[str, List[Task]]): Dictionary with the resources as
            keys and the tasks ran on them as values.
        intended_replicas (int): Number of replicas to run on each resource.
    Returns:
        Dict[str, int]: Dictionary with the resources as keys and the number of
            tasks that need to be run on them as values.
        Example: { "n1-standard-4": 2, "n1-standard-8": 1}
    """
    already_ran = defaultdict(list, already_ran)
    return {m: intended_replicas - len(already_ran[m]) for m in machines_to_run}


def _print_info(already_ran: Dict[str, List[Task]], tasks_to_run: Dict[str,
                                                                       int]):
    """
    Prints the information about the tasks that have already been run and the
    tasks that need to be run.
    """

    print()
    print(__("tasks_ran"))
    for key, value in already_ran.items():
        print(f"--->{key}: {len(value)}")
    print()
    print(__("tasks_to_run"))
    for key, value in tasks_to_run.items():
        print(f"--->{key}: {value}")
    print()


def _run_replica(simulator_args: Dict[str, Union[Callable, Any]],\
                 input_args: Dict[str, Union[Callable,Any]],\
                 resource: BaseMachineGroup,\
                 simulator: Simulator,\
                 input_files: str,\
                 replica: int):
    """Runs a replica of the benchmark.

    Args:
        simulator_args (Dict[str, Union[Callable, Any]]): Arguments for the
            simulator.
        input_args (Dict[str, Union[Callable, Any]]): Arguments for the
            input files.
        resource (BaseMachineGroup): Resource to run the benchmark on.
        simulator (Simulator): Simulator to run.
        input_files (str): Path to the input files.
        replica (int): Number of the replica.
    """
    # Each replica can run with different arguments
    simulator_args_current = _render_dict(simulator_args, resource)
    input_args_current = _render_dict(input_args, resource)
    print(f"Simulation arguments {simulator_args_current}")
    # input_files either come from the template manager or
    # are passed directly.
    simulator_args_current["input_dir"] = input_files
    simulator_args_current["on"] = resource

    if input_args_current:
        target_dir = f"benchmark_inputs/{resource.machine_type}_{replica}"
        TemplateManager.render_dir(source_dir=input_files,
                                   target_dir=target_dir,
                                   overwrite=False,
                                   **input_args)
        simulator_args_current["input_dir"] = target_dir

    _ = simulator.run(**simulator_args_current)


def _run(project: Project,
         input_files: str,
         replicas: int,
         machines: List[str],
         simulator: Simulator,
         input_args: Dict[str, Union[Callable, Any]] = None,
         machine_args: Dict[str, Union[Callable, Any]] = None,
         machine_class: Type[BaseMachineGroup] = inductiva.resources.
         MachineGroup,
         simulator_args: Dict[str, Union[Callable, Any]] = None):
    """Runs a benchmark.
    This method runs a benchmark using the provided parameters.
    Args:
        project (Project): Project to run the benchmark on.
    """
    # create dict with resource type: list of tasks ran on that resource
    # Example
    # { "n1-standard-4": [task1, task2], "n1-standard-8": [task3]}
    current_project_tasks = _tasks_by_vm_type(project)
    to_run = _compute_tasks_to_run(machines, current_project_tasks, replicas)

    print()
    if not to_run:
        print(__("all_tasks_run"))
        project.close()
        return

    _print_info(current_project_tasks, to_run)

    for machine, replicas_to_run in to_run.items():

        machine_args_current = _render_dict(machine_args, machine)

        print(f"\nCreating machine group {machine} with {machine_args_current}")

        resource = machine_class(machine_type=machine,
                                 max_idle_time=datetime.timedelta(minutes=1),
                                 **machine_args_current)

        print(f"\nRunning {simulator}")
        if resource is not None:
            while not _can_start_resource(resource):
                print("This machine will exceed the current quotas.\n"
                      "Going to sleep and trying again later.")
                time.sleep(QUOTAS_EXCEEDED_SLEEP_TIME)

            resource.start()

        for replica in range(replicas_to_run, 0, -1):
            _run_replica(simulator_args, input_args, resource, simulator,
                         input_files, replica)


def run(name: str,
        input_files: str,
        replicas: int,
        machines: List[str],
        simulator: Simulator,
        append: bool = False,
        input_args: Dict[str, Union[Callable, Any]] = None,
        machine_class: Type[BaseMachineGroup] = MachineGroup,
        machine_args: Dict[str, Union[Callable, Any]] = None,
        simulator_args: Dict[str, Union[Callable, Any]] = None):
    """
    Runs a benchmark.

    This method creates a project and runs a benchmark using the provided
    parameters. It checks if a benchmark with the given name already exists. If
    it does, the user is prompted to either append tasks to the existing
    benchmark or cancel the operation. The method waits for quotas to be
    available before starting the necessary resources. If the benchmark is
    already completed, no further action is taken. If the benchmark is
    incomplete, only the remaining tasks are executed.

    input_args, machine_args and simulator_args can be dictionaries where the
    values can be callables that will be called with the resource object as
    parameter (or str machine_type when it comes to machine_args). This is
    useful when you want to calculate the values based on the machine_group or
    machine_type.
    Example:
        def data_disk_gb(name):
            #makes the data disk size relative to the number of vcpus
            vcpu = int(name.split("-")[2])
            return vcpu+10
        machine_args={"data_disk_gb": data_disk_gb}

    Args:
            name (str): Name of the benchmark.
            input_files (str): Path to the input files.
            replicas (int): Number of replicas to run on each resource.
            machines (List[str]): List of machines to run the simulation on.
                The machine in this case is just a string with the machine type.
                Ex: ["c2-standard-4", "c2d-standard-4"]
            simulator (Simulator): Simulator to run.
            append (bool): If True, appends the tasks to a benchmark regardless
            if that benchmark is new or not. If False, will prompt the user if
            the benchmark already exists.
            machine_class (BaseMachineGroup): Class to use to create the machine
                group. If not given, MachineGroup will be used.
            input_args (Dict[str, Union[Callable, Any]]): Arguments for the
                input files.
            machine_args (Dict[str, Union[Callable, Any]]): Arguments for the
                machines.
            simulator_args (Dict[str, Union[Callable, Any]]): Arguments for the
                simulator.
    """

    project = Project(name, append=True)
    if project.num_tasks > 0 and not append:
        raise RuntimeError(
            __("benchmark-already-exists", name, project.num_tasks))

    with project:
        _run(project, input_files, replicas, machines, simulator, input_args,
             machine_args, machine_class, simulator_args)


def _render_dict(dictionary: Dict[str, Union[Callable, Any]], argument: Any):
    """Replaces callables from a dictionary.
    This function replaces callables with the result of calling the callable
    with the argument passed.
    Args:
        dictionary (Dict[str, Union[Callable, Any]]): Dictionary to replace
            callables from.
        argument: Argument to pass to the callables.
    Returns:
        Dict[str, Any]: Dictionary with callables replaced.
    """
    if dictionary:
        return {
            k: v(argument) if callable(v) else v for k, v in dictionary.items()
        }
    return {}


def _can_start_resource(resource: BaseMachineGroup) -> bool:
    """Check if the resource can be started.

        This method checks if the resource can be started by checking
        the available quotas and resource usage.

        returns:
            bool: True if the resource can be started, False otherwise.
        """
    quotas = users.get_quotas()

    cost_in_use = quotas["max_price_hour"]["in_use"]
    vcpu_in_use = quotas["max_vcpus"]["in_use"]
    machines_in_use = quotas["max_instances"]["in_use"]

    cost_max = quotas["max_price_hour"]["max_allowed"]
    vcpu_max = quotas["max_vcpus"]["max_allowed"]
    machines_max = quotas["max_instances"]["max_allowed"]

    current_vcpu = resource.n_vcpus.total

    estimated_cost = cost_in_use + resource.estimate_cloud_cost(verbose=False)
    estimated_vcpu_usage = vcpu_in_use + current_vcpu
    estimated_machine_usage = machines_in_use + 1
    return estimated_cost <= cost_max and \
            estimated_vcpu_usage <= vcpu_max and \
            estimated_machine_usage <= machines_max



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ExecuterTrackerRegisterInfo(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Info for creating an executer.
    """


    class MetaOapg:
        required = {
            "vm_id",
            "memory",
            "create_time",
            "cpu_count_logical",
            "cpu_count_physical",
            "machine_group_id",
            "vm_name",
        }
        
        class properties:
            machine_group_id = schemas.UUIDSchema
            create_time = schemas.DateTimeSchema
            cpu_count_logical = schemas.IntSchema
            cpu_count_physical = schemas.IntSchema
            memory = schemas.IntSchema
            vm_name = schemas.StrSchema
            vm_id = schemas.StrSchema
            mpi_cluster = schemas.BoolSchema
            __annotations__ = {
                "machine_group_id": machine_group_id,
                "create_time": create_time,
                "cpu_count_logical": cpu_count_logical,
                "cpu_count_physical": cpu_count_physical,
                "memory": memory,
                "vm_name": vm_name,
                "vm_id": vm_id,
                "mpi_cluster": mpi_cluster,
            }
    
    vm_id: MetaOapg.properties.vm_id
    memory: MetaOapg.properties.memory
    create_time: MetaOapg.properties.create_time
    cpu_count_logical: MetaOapg.properties.cpu_count_logical
    cpu_count_physical: MetaOapg.properties.cpu_count_physical
    machine_group_id: MetaOapg.properties.machine_group_id
    vm_name: MetaOapg.properties.vm_name
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["machine_group_id"]) -> MetaOapg.properties.machine_group_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["create_time"]) -> MetaOapg.properties.create_time: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["cpu_count_logical"]) -> MetaOapg.properties.cpu_count_logical: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["cpu_count_physical"]) -> MetaOapg.properties.cpu_count_physical: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["memory"]) -> MetaOapg.properties.memory: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["vm_name"]) -> MetaOapg.properties.vm_name: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["vm_id"]) -> MetaOapg.properties.vm_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["mpi_cluster"]) -> MetaOapg.properties.mpi_cluster: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["machine_group_id", "create_time", "cpu_count_logical", "cpu_count_physical", "memory", "vm_name", "vm_id", "mpi_cluster", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["machine_group_id"]) -> MetaOapg.properties.machine_group_id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["create_time"]) -> MetaOapg.properties.create_time: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["cpu_count_logical"]) -> MetaOapg.properties.cpu_count_logical: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["cpu_count_physical"]) -> MetaOapg.properties.cpu_count_physical: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["memory"]) -> MetaOapg.properties.memory: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["vm_name"]) -> MetaOapg.properties.vm_name: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["vm_id"]) -> MetaOapg.properties.vm_id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["mpi_cluster"]) -> typing.Union[MetaOapg.properties.mpi_cluster, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["machine_group_id", "create_time", "cpu_count_logical", "cpu_count_physical", "memory", "vm_name", "vm_id", "mpi_cluster", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        vm_id: typing.Union[MetaOapg.properties.vm_id, str, ],
        memory: typing.Union[MetaOapg.properties.memory, decimal.Decimal, int, ],
        create_time: typing.Union[MetaOapg.properties.create_time, str, datetime, ],
        cpu_count_logical: typing.Union[MetaOapg.properties.cpu_count_logical, decimal.Decimal, int, ],
        cpu_count_physical: typing.Union[MetaOapg.properties.cpu_count_physical, decimal.Decimal, int, ],
        machine_group_id: typing.Union[MetaOapg.properties.machine_group_id, str, uuid.UUID, ],
        vm_name: typing.Union[MetaOapg.properties.vm_name, str, ],
        mpi_cluster: typing.Union[MetaOapg.properties.mpi_cluster, bool, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'ExecuterTrackerRegisterInfo':
        return super().__new__(
            cls,
            *_args,
            vm_id=vm_id,
            memory=memory,
            create_time=create_time,
            cpu_count_logical=cpu_count_logical,
            cpu_count_physical=cpu_count_physical,
            machine_group_id=machine_group_id,
            vm_name=vm_name,
            mpi_cluster=mpi_cluster,
            _configuration=_configuration,
            **kwargs,
        )



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class MachineType(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "machine_type",
            "num_cpus",
            "price",
            "ram_gb",
        }
        
        class properties:
            machine_type = schemas.StrSchema
            num_cpus = schemas.IntSchema
            ram_gb = schemas.IntSchema
            price = schemas.NumberSchema
            
            
            class provider_id(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    
                    @classmethod
                    @functools.lru_cache()
                    def all_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            Providers,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'provider_id':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class threads_per_core(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.IntSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'threads_per_core':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class spot(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.BoolSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'spot':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class region(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'region':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class zone(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'zone':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            __annotations__ = {
                "machine_type": machine_type,
                "num_cpus": num_cpus,
                "ram_gb": ram_gb,
                "price": price,
                "provider_id": provider_id,
                "threads_per_core": threads_per_core,
                "spot": spot,
                "region": region,
                "zone": zone,
            }
    
    machine_type: MetaOapg.properties.machine_type
    num_cpus: MetaOapg.properties.num_cpus
    price: MetaOapg.properties.price
    ram_gb: MetaOapg.properties.ram_gb
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["machine_type"]) -> MetaOapg.properties.machine_type: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["num_cpus"]) -> MetaOapg.properties.num_cpus: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["ram_gb"]) -> MetaOapg.properties.ram_gb: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["price"]) -> MetaOapg.properties.price: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["provider_id"]) -> MetaOapg.properties.provider_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["threads_per_core"]) -> MetaOapg.properties.threads_per_core: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["spot"]) -> MetaOapg.properties.spot: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["region"]) -> MetaOapg.properties.region: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["zone"]) -> MetaOapg.properties.zone: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["machine_type", "num_cpus", "ram_gb", "price", "provider_id", "threads_per_core", "spot", "region", "zone", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["machine_type"]) -> MetaOapg.properties.machine_type: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["num_cpus"]) -> MetaOapg.properties.num_cpus: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["ram_gb"]) -> MetaOapg.properties.ram_gb: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["price"]) -> MetaOapg.properties.price: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["provider_id"]) -> typing.Union[MetaOapg.properties.provider_id, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["threads_per_core"]) -> typing.Union[MetaOapg.properties.threads_per_core, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["spot"]) -> typing.Union[MetaOapg.properties.spot, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["region"]) -> typing.Union[MetaOapg.properties.region, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["zone"]) -> typing.Union[MetaOapg.properties.zone, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["machine_type", "num_cpus", "ram_gb", "price", "provider_id", "threads_per_core", "spot", "region", "zone", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        machine_type: typing.Union[MetaOapg.properties.machine_type, str, ],
        num_cpus: typing.Union[MetaOapg.properties.num_cpus, decimal.Decimal, int, ],
        price: typing.Union[MetaOapg.properties.price, decimal.Decimal, int, float, ],
        ram_gb: typing.Union[MetaOapg.properties.ram_gb, decimal.Decimal, int, ],
        provider_id: typing.Union[MetaOapg.properties.provider_id, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        threads_per_core: typing.Union[MetaOapg.properties.threads_per_core, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        spot: typing.Union[MetaOapg.properties.spot, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        region: typing.Union[MetaOapg.properties.region, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        zone: typing.Union[MetaOapg.properties.zone, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'MachineType':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            num_cpus=num_cpus,
            price=price,
            ram_gb=ram_gb,
            provider_id=provider_id,
            threads_per_core=threads_per_core,
            spot=spot,
            region=region,
            zone=zone,
            _configuration=_configuration,
            **kwargs,
        )

from inductiva.client.model.providers import Providers



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class ExecuterTrackerAPIConnectionInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Information sent to the executer tracker after registration.
    """

    class MetaOapg:
        required = {
            "redis_consumer_group",
            "redis_stream",
            "redis_consumer_name",
            "executer_tracker_id",
            "machine_group_id",
        }

        class properties:
            executer_tracker_id = schemas.UUIDSchema
            redis_stream = schemas.StrSchema
            redis_consumer_name = schemas.StrSchema
            redis_consumer_group = schemas.StrSchema
            machine_group_id = schemas.UUIDSchema
            __annotations__ = {
                "executer_tracker_id": executer_tracker_id,
                "redis_stream": redis_stream,
                "redis_consumer_name": redis_consumer_name,
                "redis_consumer_group": redis_consumer_group,
                "machine_group_id": machine_group_id,
            }

    redis_consumer_group: MetaOapg.properties.redis_consumer_group
    redis_stream: MetaOapg.properties.redis_stream
    redis_consumer_name: MetaOapg.properties.redis_consumer_name
    executer_tracker_id: MetaOapg.properties.executer_tracker_id
    machine_group_id: MetaOapg.properties.machine_group_id

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["executer_tracker_id"]
    ) -> MetaOapg.properties.executer_tracker_id:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["redis_stream"]
    ) -> MetaOapg.properties.redis_stream:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["redis_consumer_name"]
    ) -> MetaOapg.properties.redis_consumer_name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["redis_consumer_group"]
    ) -> MetaOapg.properties.redis_consumer_group:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_group_id"]
    ) -> MetaOapg.properties.machine_group_id:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "executer_tracker_id",
        "redis_stream",
        "redis_consumer_name",
        "redis_consumer_group",
        "machine_group_id",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["executer_tracker_id"]
    ) -> MetaOapg.properties.executer_tracker_id:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["redis_stream"]
    ) -> MetaOapg.properties.redis_stream:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["redis_consumer_name"]
    ) -> MetaOapg.properties.redis_consumer_name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["redis_consumer_group"]
    ) -> MetaOapg.properties.redis_consumer_group:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_group_id"]
    ) -> MetaOapg.properties.machine_group_id:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "executer_tracker_id",
        "redis_stream",
        "redis_consumer_name",
        "redis_consumer_group",
        "machine_group_id",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        redis_consumer_group: typing.Union[
            MetaOapg.properties.redis_consumer_group,
            str,
        ],
        redis_stream: typing.Union[
            MetaOapg.properties.redis_stream,
            str,
        ],
        redis_consumer_name: typing.Union[
            MetaOapg.properties.redis_consumer_name,
            str,
        ],
        executer_tracker_id: typing.Union[
            MetaOapg.properties.executer_tracker_id,
            str,
            uuid.UUID,
        ],
        machine_group_id: typing.Union[
            MetaOapg.properties.machine_group_id,
            str,
            uuid.UUID,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'ExecuterTrackerAPIConnectionInfo':
        return super().__new__(
            cls,
            *_args,
            redis_consumer_group=redis_consumer_group,
            redis_stream=redis_stream,
            redis_consumer_name=redis_consumer_name,
            executer_tracker_id=executer_tracker_id,
            machine_group_id=machine_group_id,
            _configuration=_configuration,
            **kwargs,
        )



"""Methods to interact with the user storage resources."""
import inductiva
from inductiva.client.apis.tags import storage_api
from inductiva.utils import format_utils
from typing import Literal


def get_space_used():
    """Returns the occupied storage size in GB."""
    try:
        api = storage_api.StorageApi(inductiva.api.get_client())
        response = api.get_storage_size()
        storage_used = round(float(response.body) / (1024**3), 3)
        print(f"Total user's remote storage in use: {storage_used} GB")
        return storage_used
    except inductiva.client.ApiException as api_exception:
        raise api_exception


def listdir(path="/",
            max_results: int = 10,
            order_by: Literal["size", "creation_time"] = "creation_time",
            sort_order: Literal["asc", "desc"] = "desc"):
    """List and display the contents of the user's storage.
    Args:
        path (str): Storage directory to list. Default is root.
        max_results (int): The maximum number of results to return.
        order_by (str): The field to sort the contents by.
        sort_order (str): Whether to sort the contents in ascending or
        descending order.
    Returns:
        list of dict: A list of dictionaries containing information about
        the size, the name and the creation time of each content that can
        easily be converted to a dataframe.

    This function prints a table with the storage content information:
        Name            Size            Creation Time
        1234            5.68 MiB        29 Sep, 14:12:00
        12345           374.85 KiB      29 Sep, 14:13:10
        1234567         97.59 KiB       29 Sep, 14:13:24
        123             0 B             29 Sep, 14:13:29
        123456          0 B             29 Sep, 14:13:18
    You can use this information to delete the contents you don't need
    anymore and further inspect task outputs and logs using the Task
    class.
    """

    api = storage_api.StorageApi(inductiva.api.get_client())
    contents = api.list_storage_contents({
        "path": path,
        "max_results": max_results,
        "sort_by": order_by,
        "order": sort_order
    }).body
    all_contents = []
    for content_name, info in contents.items():
        size = info["size_bytes"]
        creation_time = info["creation_time"]
        all_contents.append({
            "content_name": content_name,
            "size": round(float(size), 3),
            "creation_time": creation_time
        })
    print(_print_contents_table(all_contents))
    print(f"Listed {len(all_contents)} folder(s). Ordered by {order_by}.\n"
          "Use --max-results/-m to control the number of results displayed.")
    return all_contents


def _print_contents_table(contents):
    columns = ["Name", "Size", "Creation Time"]
    rows = []

    for content in contents:
        row = [
            content["content_name"], content["size"], content["creation_time"]
        ]
        rows.append(row)

    formatters = {
        "Creation Time": [format_utils.datetime_formatter],
        "Size": [format_utils.bytes_formatter],
    }

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    return format_utils.get_tabular_str(
        rows,
        columns,
        formatters=formatters,
        header_formatters=header_formatters,
    )



"""List the available simulators via CLI."""
import argparse

from inductiva.simulators import methods
from inductiva.utils import format_utils


def list_versions(args):
    """List available simulators and associated versions for each branch."""

    available = methods.list_available_images()

    if not args.dev:
        available.pop("development")
    fmtr = format_utils.get_ansi_formatter()

    def header_fmtr(s):
        return fmtr(s.upper(), format_utils.Emphasis.BOLD)

    for branch, simulators in available.items():
        sims = sorted(simulators.keys())

        data = {
            "simulator": sims,
            "versions": [", ".join(simulators[s]) for s in sims],
        }

        table = format_utils.get_tabular_str(data,
                                             header_formatters=[header_fmtr])
        header = f"Available simulators and versions for {branch} runs:"

        print(header_fmtr(header))
        print(table)


def register(parser):
    """Register the "simulators list" sub-command."""

    subparser = parser.add_parser("list",
                                  aliases=["ls"],
                                  help="List available simulators.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `list` sub-command lists all available simulators and\n"
        "associated versions for use with the Inductiva API,\n"
        "including those available for development purposes.\n")
    subparser.add_argument(
        "--dev",
        action="store_true",
        help="include development versions of the simulators.")

    subparser.set_defaults(func=list_versions)



"""List the user projects information via CLI."""
from collections import defaultdict
from typing import TextIO
import argparse
import sys

import inductiva
from inductiva import _cli
from inductiva.utils import format_utils


def get_projects(_, fout: TextIO = sys.stdout):
    """Lists all the user's projects."""

    projects = inductiva.projects.project.get_projects()
    table = defaultdict(list)
    for p in projects:
        table["name"].append(p.name)
        table["nr_tasks"].append(p.num_tasks)

    emph_formatter = format_utils.get_ansi_formatter()

    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    table = format_utils.get_tabular_str(table,
                                         header_formatters=header_formatters)

    print(table, file=fout, end="")

    return 0


def register(parser):
    """Register the projects list command."""

    subparser = parser.add_parser("list",
                                  aliases=["ls"],
                                  help="List the user's projects.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva projects list` command provides "
                             "an overview of your projects.\n"
                             "It lists all your available projects.")

    _cli.utils.add_watch_argument(subparser)

    subparser.set_defaults(func=get_projects)



"""Project class"""
import contextvars
import logging
from typing import Optional, Union

import inductiva
from inductiva.client import ApiException
from inductiva.client import models
from inductiva.client.apis.tags import projects_api
from inductiva.client.model.project import Project as ProjectModel

_logger = logging.getLogger(__name__)

_CURRENT_PROJECT = contextvars.ContextVar("current_project", default=None)


def get_current_project():
    """Get the current active project.
    
    Returns the currently active project in the calling thread.  If no
    project has been defined, the default project (`None`) will be
    returned.

    """
    return _CURRENT_PROJECT.get()


class ProjectInfo:
    """Stores info about the project."""

    def __init__(self, **attrs):
        self.task_by_status = {
            models.TaskStatusCode(attr): int(value)
            for attr, value in attrs.items()
        }


def get_projects():
    """Gets all the user's projects."""
    try:
        _logger.debug("Trying to get remote projects")
        api = projects_api.ProjectsApi(inductiva.api.get_client())
        response = api.get_user_projects()
    except ApiException as ex:
        _logger.error("Failed to get remote projects", exc_info=ex)
        raise ex

    return [Project.from_api_response(resp) for resp in response.body]


class Project:
    """Projects management class.
 
    The Project class allows tasks to be collected under the umbrella
    of a single container object. All tasks submitted in the context
    of an opened project will be associated with that project and can
    be retrieved altogether using the project as a filter. A project
    can be created and managed declaratively or using a context
    manager.
 
    Example usage:
 
    >>> project = inductiva.projects.Project("test_project")
    >>> project.start()
    >>> simulator.run(...)  # Submitted to  `'test_project'`
    >>> project.stop()
 
    This is equivalent to:
 
    >>> with inductiva.projects.Project("test_project"):
    >>>    simulator.run(...)

    """

    def __init__(self, name: str, *, append: bool = False):
        """Initialize the Project instance.
        
        Args:
          name (str): The name of the project.
          append (bool): A flag indicating that new tasks can be appended to
              the project. If set to False (default), task submission will fail,
              even though the project might be opened.
        """
        self.append = append
        self._token = None

        model = self._get_model(name)
        if model is None:
            model = self._create_model(name)
        self._update_from_api_response(model)

    @classmethod
    def from_api_response(cls, model: ProjectModel) -> "Project":
        """Builds a `Project` object from an API response.

        Args:
          model: Info about the project. This should be fetched from the
          backend.

        """
        project = cls.__new__(Project)
        project.append = False
        project._token = None
        return project._update_from_api_response(model)

    @staticmethod
    def _get_model(name: str) -> ProjectModel:
        """Fetches the project info from the backend.

        Args:
          name: The name of the project

        Returns:
          Project models object with the info about the project.

        """
        try:
            _logger.debug("Trying to get remote project %s", name)
            api = projects_api.ProjectsApi(inductiva.api.get_client())
            response = api.get_project({"name": name})
            return response.body
        except ApiException as ex:
            if ex.status != 404:
                _logger.error("Failed to get remote project %s",
                              name,
                              exc_info=ex)
                raise ex
            _logger.debug("Project %s does not exist", name)
        return None

    @staticmethod
    def _create_model(name):
        """Creates a project with the given name on the backend."""
        try:
            _logger.debug("Creating remote project %s", name)
            api = projects_api.ProjectsApi(inductiva.api.get_client())
            response = api.create_project({"name": name})
            return response.body
        except ApiException as ex:
            _logger.error("Failed to create remote project %s",
                          name,
                          exc_info=ex)
            raise RuntimeError(f"Unable to create project {name}") from ex

    def _update_from_api_response(self, model: ProjectModel) -> "Project":
        """Updates itself with information fetched from the backend."""
        self._info = ProjectInfo(**model.get("task_status_overview"))
        self.created_at = model.get("created_at")
        self.num_tasks = int(model.get("num_tasks"))
        self._name = model.get("name")
        self.id = model.get("id")

        return self

    def open(self):
        """Open the project.

        Open the project and make it the active one in the calling
        thread. An opened project will ensure that calls to the
        `get_current_project` function will return this project.
        Consecutive calls to this method are idempotent.

        Raises:
            RuntimeError if another opened project exists, i.e.,
            `get_current_project` returns something other than None.

        """
        if not self.append:
            raise RuntimeError(
                "Trying to open a project with `append=False`.\n"
                "A Project can only be opened when instantiated with the"
                " `append=True` option.")

        _logger.debug("Opening project %s", self._name)
        current_project = get_current_project()

        if current_project is self:
            _logger.debug("Project is already opened.")
            return

        if current_project is not None:
            raise RuntimeError(
                "Trying to open a project when another is active.")

        self._token = _CURRENT_PROJECT.set(self)

    def close(self):
        """Close the project.

        Calls to the `get_current_project` will return `None` after
        the project is closed.  Consecutive calls to this method are
        idempotent.

        """
        if self._token is None:
            return

        _CURRENT_PROJECT.reset(self._token)
        self._token = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def opened(self) -> bool:
        """Checks if the project is open."""
        return self._token is not None

    @property
    def info(self) -> ProjectInfo:
        """Returns project info.

        It makes no requests to the backend. Instead it uses
        previously stored information. Therefore it can be
        outdated. For a method that requests the most recent data from
        the backend use: `get_info`.

        """
        return self._info

    def get_info(self) -> ProjectInfo:
        """Get project information.

        Get updated information on the project. This method executes a
        call to the backend to retrieve the most recent information
        about this project and stores it internally so that it becomes
        available through the `Project.info` property.  This method is
        suitable when up-to-date information is required.

        """
        name = self._name
        model = self._get_model(name)
        self._update_from_api_response(model)
        return self._info

    def __str__(self) -> str:
        return f"Project '{self._name}' with "\
               f"{self.num_tasks} tasks (id={self.id})"

    def describe(self) -> str:
        """Generates a string description of the object

        Returns:
          str: A string description of the object. Includes project name,
            total number of tasks and the number of tasks by status.

        """
        header = str(self) + "\n"
        summary = "\n".join(
            f"  {k}: {v}" for k, v in self._info.task_by_status.items())
        return header + summary

    def get_tasks(self,
                  last_n: int = -1,
                  status: Optional[Union[str, models.TaskStatusCode]] = None):
        """Get the last N submitted tasks to this project.

        Get the last N submitted tasks that belong to this project,
        eventually filtered by status. By default, only the last 5
        submitted tasks are returned, irrespectively of their status.

        Args:
            last_n (int): The number of tasks with repect to the submission
                time to fectch. If `last_n<=0` we fetch all tasks submitted
                to the project.
            status: Status of the tasks to get. If `None`, tasks with any
                status will be returned.
        """
        return inductiva.tasks.get_tasks(last_n=last_n,
                                         project=self,
                                         status=status)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def __eq__(self, other) -> bool:
        return isinstance(other, Project) and \
            self.name == other.name and \
            self.id == other.id



"""Functions to manage or retrieve user resources."""
from typing import Optional
import logging

import inductiva
import inductiva.client
from inductiva.client.apis.tags import compute_api
from inductiva import resources


def estimate_machine_cost(machine_type: str, spot: bool = False):
    """Estimate the cloud cost of one machine per hour in US dollars.

    Args:
        machine_type: The type of GC machine to launch. Ex: "e2-standard-4".
            Check https://cloud.google.com/compute/docs/machine-resource for
            more information about machine types.
        zone: The zone where the machines will be launched.
        spot: Whether to use spot machines.
    """

    api = compute_api.ComputeApi(inductiva.api.get_client())

    instance_price = api.get_instance_price({
        "machine_type": machine_type,
    })

    if spot:
        estimated_cost = instance_price.body["preemptible_price"]
    else:
        estimated_cost = instance_price.body["on_demand_price"]

    return float(estimated_cost)


def _fetch_machine_groups_from_api():
    """Get all active machine groups of a user from the API."""
    try:
        api = compute_api.ComputeApi(inductiva.api.get_client())
        response = api.list_active_user_instance_groups()

        return response.body

    except inductiva.client.ApiException as api_exception:
        raise api_exception


def get_by_name(machine_name: str):
    """Returns the machine group corresponding to `machine_name`."""
    try:
        api = compute_api.ComputeApi(inductiva.api.get_client())
        response = api.get_vm_group_by_name({"name": machine_name}).body
        mg_class = _get_machine_group_class(response["type"],
                                            response["is_elastic"])
        return mg_class.from_api_response(response)
    except inductiva.client.ApiException as api_exception:
        raise api_exception


def _get_machine_group_class(machine_type: str, is_elastic: bool):
    """Returns the class of the machine group"""
    if is_elastic:
        mg_class = resources.ElasticMachineGroup
    elif machine_type == "standard":
        mg_class = resources.MachineGroup
    elif machine_type == "mpi":
        mg_class = resources.MPICluster
    else:
        raise ValueError("Unknown resource configuration.")
    return mg_class


def get():
    """Returns a list of 'Resource' objects."""

    # Retrive the active resource names
    machine_groups = _fetch_machine_groups_from_api()
    machine_group_list = []

    for mg in machine_groups:
        mg_class = _get_machine_group_class(mg["type"], mg["is_elastic"])
        machine_group_list.append(mg_class.from_api_response(mg))

    return machine_group_list


def get_cheapest_machine_type(num_cpus: int,
                              ram_gb: Optional[int] = None,
                              spot: bool = True):
    """Get the machine type with the lowest price.

    If the machine type with the exact requirements is not found, the closest
    machine type with higher resources will be returned. For example, if a
    machine type with 5 CPUs and 20 GB of RAM is requested, but the closest
    machine type has 8 CPUs and 32 GB of RAM, the latter will be returned.

    Args:
        num_cpus: The minimum number of CPUs required.
        ram_gb: The minimum RAM in GB required. If None, the minimum RAM for
          the given number of CPUs will be used.
        spot: Whether the machine is spot or not.
    """
    spot = "t" if spot else "f"
    api = compute_api.ComputeApi(inductiva.api.get_client())
    body = {"num_cpus": num_cpus, "spot": spot}
    if ram_gb:
        body["ram_gb"] = ram_gb
    response = api.get_machine_type(body)

    logging.info("Machine type: %s", response.body["machine_type"])
    logging.info("CPUs: %s", response.body["num_cpus"])
    logging.info("RAM in GB: %s", response.body["ram_gb"])
    logging.info("Estimated cost per hour: %s $/h",
                 round(response.body["price"], 5))

    return response.body["machine_type"]



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class OutputArchiveInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "size",
            "contents",
        }

        class properties:
            size = schemas.IntSchema

            class contents(schemas.ListSchema):

                class MetaOapg:

                    @staticmethod
                    def items() -> typing.Type['FileInfo']:
                        return FileInfo

                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['FileInfo'],
                                       typing.List['FileInfo']],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                ) -> 'contents':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )

                def __getitem__(self, i: int) -> 'FileInfo':
                    return super().__getitem__(i)

            __annotations__ = {
                "size": size,
                "contents": contents,
            }

    size: MetaOapg.properties.size
    contents: MetaOapg.properties.contents

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["size"]
    ) -> MetaOapg.properties.size:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["contents"]
    ) -> MetaOapg.properties.contents:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "size",
        "contents",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["size"]
    ) -> MetaOapg.properties.size:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["contents"]
    ) -> MetaOapg.properties.contents:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "size",
        "contents",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        size: typing.Union[
            MetaOapg.properties.size,
            decimal.Decimal,
            int,
        ],
        contents: typing.Union[
            MetaOapg.properties.contents,
            list,
            tuple,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'OutputArchiveInfo':
        return super().__new__(
            cls,
            *_args,
            size=size,
            contents=contents,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.file_info import FileInfo



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskRequest(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "method",
        }
        
        class properties:
            method = schemas.StrSchema
            
            
            class resource_pool(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    format = 'uuid4'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'resource_pool':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            storage_path_prefix = schemas.StrSchema
            
            
            class provider_id(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    
                    @classmethod
                    @functools.lru_cache()
                    def all_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            Providers,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'provider_id':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class project(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'project':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class container_image(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'container_image':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            
            
            class time_to_live_seconds(
                schemas.ComposedSchema,
            ):
            
            
                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema
                    
                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]
            
            
                def __new__(
                    cls,
                    *_args: typing.Union[dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, ],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
                ) -> 'time_to_live_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )
            resubmit_on_preemption = schemas.BoolSchema
            __annotations__ = {
                "method": method,
                "resource_pool": resource_pool,
                "storage_path_prefix": storage_path_prefix,
                "provider_id": provider_id,
                "project": project,
                "container_image": container_image,
                "time_to_live_seconds": time_to_live_seconds,
                "resubmit_on_preemption": resubmit_on_preemption,
            }
    
    method: MetaOapg.properties.method
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["method"]) -> MetaOapg.properties.method: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["resource_pool"]) -> MetaOapg.properties.resource_pool: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["storage_path_prefix"]) -> MetaOapg.properties.storage_path_prefix: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["provider_id"]) -> MetaOapg.properties.provider_id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["project"]) -> MetaOapg.properties.project: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["container_image"]) -> MetaOapg.properties.container_image: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["time_to_live_seconds"]) -> MetaOapg.properties.time_to_live_seconds: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["resubmit_on_preemption"]) -> MetaOapg.properties.resubmit_on_preemption: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["method", "resource_pool", "storage_path_prefix", "provider_id", "project", "container_image", "time_to_live_seconds", "resubmit_on_preemption", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["method"]) -> MetaOapg.properties.method: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["resource_pool"]) -> typing.Union[MetaOapg.properties.resource_pool, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["storage_path_prefix"]) -> typing.Union[MetaOapg.properties.storage_path_prefix, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["provider_id"]) -> typing.Union[MetaOapg.properties.provider_id, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["project"]) -> typing.Union[MetaOapg.properties.project, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["container_image"]) -> typing.Union[MetaOapg.properties.container_image, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["time_to_live_seconds"]) -> typing.Union[MetaOapg.properties.time_to_live_seconds, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["resubmit_on_preemption"]) -> typing.Union[MetaOapg.properties.resubmit_on_preemption, schemas.Unset]: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["method", "resource_pool", "storage_path_prefix", "provider_id", "project", "container_image", "time_to_live_seconds", "resubmit_on_preemption", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        method: typing.Union[MetaOapg.properties.method, str, ],
        resource_pool: typing.Union[MetaOapg.properties.resource_pool, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        storage_path_prefix: typing.Union[MetaOapg.properties.storage_path_prefix, str, schemas.Unset] = schemas.unset,
        provider_id: typing.Union[MetaOapg.properties.provider_id, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        project: typing.Union[MetaOapg.properties.project, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        container_image: typing.Union[MetaOapg.properties.container_image, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        time_to_live_seconds: typing.Union[MetaOapg.properties.time_to_live_seconds, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO, io.BufferedReader, schemas.Unset] = schemas.unset,
        resubmit_on_preemption: typing.Union[MetaOapg.properties.resubmit_on_preemption, bool, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'TaskRequest':
        return super().__new__(
            cls,
            *_args,
            method=method,
            resource_pool=resource_pool,
            storage_path_prefix=storage_path_prefix,
            provider_id=provider_id,
            project=project,
            container_image=container_image,
            time_to_live_seconds=time_to_live_seconds,
            resubmit_on_preemption=resubmit_on_preemption,
            _configuration=_configuration,
            **kwargs,
        )

from inductiva.client.model.providers import Providers



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class UserMostUsedMachineTypesOverview(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "machine_type",
            "num_tasks",
            "preempted_machines",
            "num_of_machines_used",
            "spot_machines",
            "errors",
        }

        class properties:
            machine_type = schemas.StrSchema
            num_of_machines_used = schemas.IntSchema
            num_tasks = schemas.IntSchema
            spot_machines = schemas.IntSchema
            preempted_machines = schemas.IntSchema

            class errors(schemas.DictSchema):

                class MetaOapg:
                    additional_properties = schemas.IntSchema

                def __getitem__(
                        self, name: typing.Union[
                            str,
                        ]) -> MetaOapg.additional_properties:
                    # dict_instance[name] accessor
                    return super().__getitem__(name)

                def get_item_oapg(
                        self, name: typing.Union[
                            str,
                        ]) -> MetaOapg.additional_properties:
                    return super().get_item_oapg(name)

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[
                        MetaOapg.additional_properties,
                        decimal.Decimal,
                        int,
                    ],
                ) -> 'errors':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "machine_type": machine_type,
                "num_of_machines_used": num_of_machines_used,
                "num_tasks": num_tasks,
                "spot_machines": spot_machines,
                "preempted_machines": preempted_machines,
                "errors": errors,
            }

    machine_type: MetaOapg.properties.machine_type
    num_tasks: MetaOapg.properties.num_tasks
    preempted_machines: MetaOapg.properties.preempted_machines
    num_of_machines_used: MetaOapg.properties.num_of_machines_used
    spot_machines: MetaOapg.properties.spot_machines
    errors: MetaOapg.properties.errors

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_of_machines_used"]
    ) -> MetaOapg.properties.num_of_machines_used:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["num_tasks"]
    ) -> MetaOapg.properties.num_tasks:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["spot_machines"]
    ) -> MetaOapg.properties.spot_machines:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["preempted_machines"]
    ) -> MetaOapg.properties.preempted_machines:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["errors"]
    ) -> MetaOapg.properties.errors:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "machine_type",
        "num_of_machines_used",
        "num_tasks",
        "spot_machines",
        "preempted_machines",
        "errors",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_of_machines_used"]
    ) -> MetaOapg.properties.num_of_machines_used:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["num_tasks"]
    ) -> MetaOapg.properties.num_tasks:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["spot_machines"]
    ) -> MetaOapg.properties.spot_machines:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["preempted_machines"]
    ) -> MetaOapg.properties.preempted_machines:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["errors"]
    ) -> MetaOapg.properties.errors:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "machine_type",
        "num_of_machines_used",
        "num_tasks",
        "spot_machines",
        "preempted_machines",
        "errors",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        machine_type: typing.Union[
            MetaOapg.properties.machine_type,
            str,
        ],
        num_tasks: typing.Union[
            MetaOapg.properties.num_tasks,
            decimal.Decimal,
            int,
        ],
        preempted_machines: typing.Union[
            MetaOapg.properties.preempted_machines,
            decimal.Decimal,
            int,
        ],
        num_of_machines_used: typing.Union[
            MetaOapg.properties.num_of_machines_used,
            decimal.Decimal,
            int,
        ],
        spot_machines: typing.Union[
            MetaOapg.properties.spot_machines,
            decimal.Decimal,
            int,
        ],
        errors: typing.Union[
            MetaOapg.properties.errors,
            dict,
            frozendict.frozendict,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'UserMostUsedMachineTypesOverview':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            num_tasks=num_tasks,
            preempted_machines=preempted_machines,
            num_of_machines_used=num_of_machines_used,
            spot_machines=spot_machines,
            errors=errors,
            _configuration=_configuration,
            **kwargs,
        )



"""Util functions for formatting data for printing to console."""
from typing import (Any, Callable, Dict, Iterable, Mapping, Optional, Tuple,
                    Union, List)
from enum import Enum
from distutils.util import strtobool
import datetime
import os
import copy

import tabulate
from tabulate import TableFormat, DataRow

import inductiva

# pylint: disable=protected-access
tabulate._table_formats["inductiva"] = TableFormat(
    lineabove=None,
    linebelowheader=None,
    linebetweenrows=None,
    linebelow=None,
    headerrow=DataRow("", "   ", ""),
    datarow=DataRow("", "   ", ""),
    padding=0,
    with_header_hide=None,
)


class Emphasis(Enum):
    RED = "\033[31m"
    GREEN = "\033[32m"
    BOLD = "\033[1m"
    RESET = "\033[0m"


class CaseInsensitiveEnum(str, Enum):
    """Case insensitive Enum class."""

    @classmethod
    def _missing_(cls, value):
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member


def getenv_bool(varname, default):
    """Get boolean value from environment variable."""
    return bool(strtobool(os.getenv(varname, str(default))))


def no_formatter(x, *_):
    """Identity formatter, i.e, applies no formatting"""
    return x


def bytes_formatter(n_bytes: int) -> str:
    """Convert bytes to human readable string."""
    res = float(n_bytes)

    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if res < 1000:
            if unit == "B":
                return f"{res:.0f} {unit}"
            else:
                return f"{res:.2f} {unit}"
        res /= 1000

    return f"{res:.2f} PB"


def emphasis_formatter(string_to_emphasize: str, *emphasis: Emphasis):
    """Adds ansi emphasis, i.e, colors or bold to strings.

    Args:
      string_to_emphasize: String, the string to emphasize.
      emphasis: Elements of Emphasis class.

    """
    if any(emph not in [Emphasis.RED, Emphasis.GREEN, Emphasis.BOLD]
           for emph in emphasis):
        raise ValueError("Desired emphasis is not supported. "
                         "Select either `GREEN`, `RED` or `BOLD`")
    emphs = map(lambda x: x.value, emphasis)
    return "".join(emphs) + f"{string_to_emphasize}{Emphasis.RESET.value}"


def get_ansi_formatter():
    """Fetches the formatter used for ANSI formatting.

    Either `no_formatter` when ansi formatting is disable or the
    `inductiva.utils.format_utils.emphasis_formatter`
    
    """
    if not inductiva.ansi_enabled:
        return no_formatter
    return emphasis_formatter


def datetime_formatter(dt: str) -> str:
    # get time in local timezone
    if dt is None:
        return None
    local_dt = datetime.datetime.fromisoformat(dt).astimezone()
    return local_dt.strftime("%d %b, %H:%M:%S")


def datetime_formatter_ymd_hm(dt: str) -> str:
    # get time in local timezone
    if dt is None:
        return None
    local_dt = datetime.datetime.fromisoformat(dt).astimezone()
    return local_dt.strftime("%Y-%m-%d %H:%M")


def seconds_formatter(secs: float) -> str:
    """Convert seconds to time human readable string."""
    return str(datetime.timedelta(seconds=round(secs)))


def timedelta_formatter(td: datetime.timedelta) -> str:
    """Convert timedelta to human readable string."""
    parts = []
    if td.days:
        parts.append(f"{td.days} day{'s' if td.days > 1 else ''}")
    hours, remainder = divmod(td.seconds, 3600)
    if hours:
        parts.append(f"{hours} hour{'s' if hours > 1 else ''}")
    minutes, seconds = divmod(remainder, 60)
    if minutes:
        parts.append(f"{minutes} minute{'s' if minutes > 1 else ''}")
    if seconds or not parts:
        parts.append(f"{seconds} second{'s' if seconds != 1 else ''}")

    # Join parts and replace the last comma with "and"
    result = ", ".join(parts)
    parts = result.rsplit(", ", 1)
    result = " and ".join(parts) if len(parts) == 2 else result

    return result


def apply_formatters(table_data: dict, formatters: dict):
    """Applies a dict of formatters to dict of data.

    Args:
        table_data : Dictionary of column names and lists of data.
        formatters : Dictionary of column names and functions to
            apply to that column's data.
    """
    output_table_data = copy.deepcopy(table_data)
    for column_name, formatters_for_column in formatters.items():
        if column_name in output_table_data:
            for formatter in formatters_for_column:
                output_table_data[column_name] = [
                    formatter(x) for x in output_table_data[column_name]
                ]

    return output_table_data


def get_tabular_data(
    tabular_data: Union[Mapping[str, Iterable[Any]], Iterable[Iterable[Any]]],
    headers: Optional[Iterable[Any]] = None,
    formatters: Optional[Dict[str,
                              List[Callable]]] = None) -> Tuple[dict, list]:
    """Converts a table of data (Mapping or any Iterable) to
    dict and a list of headers.

    Args:
        Gets the same arguments as get_tabular_str.
    Returns:
        A dict of column names and iterable of data, and a list of headers.
    """
    formatters = formatters or {}
    headers = headers or []

    if not isinstance(tabular_data, Mapping):

        #if we have no headers data will be empty.
        #So, we want our original tabular_data
        if headers:
            tabular_data = {
                header: [row[index] for row in tabular_data]
                for index, header in enumerate(headers)
            }
    else:
        headers = list(tabular_data.keys())

    tabular_data_formatted = apply_formatters(tabular_data, formatters)
    return tabular_data_formatted, headers


def _table_indenter(table_string, num_spaces):
    "Adds spaces to the beggining of each table line."
    indentation = " " * num_spaces
    return indentation + ("\n" + indentation).join(table_string.split("\n"))


def get_tabular_str(tabular_data: Union[Mapping[str, Iterable[Any]],
                                        Iterable[Iterable[Any]]],
                    headers: Optional[Iterable[Any]] = None,
                    formatters: Optional[Dict[str, Callable]] = None,
                    header_formatters: Optional[List[Callable]] = None,
                    indentation_level: Optional[int] = 1) -> str:
    """Converts a table of data (Mapping or any Iterable) to a string table.

    Args:
        tabular_data: can be a list-of-lists (or another iterable of 
            iterables), a list of named tuples, a dictionary of
            iterables, an iterable of dictionaries, an iterable of
            dataclasses (Python 3.7+)
        headers: A list of column names.
            Only needed if tabular_data is not a Mapping.
        formatters: A dictionary of column names and functions to apply
            to the data in that column. The function should take a single
            argument and return a string. The function will be applied to the
            data in the column before printing. Defaults to None.
    Returns:
        A string table with the contents of tabular_data and headers.
    """

    data, headers = get_tabular_data(tabular_data, headers, formatters)

    header_formatters = header_formatters or []

    for formatter in header_formatters:
        headers = [formatter(header) for header in headers]

    table = tabulate.tabulate(data,
                              headers=headers,
                              missingval="n/a",
                              numalign="left",
                              tablefmt="inductiva")
    if indentation_level is not None:
        table = _table_indenter(table, indentation_level)

    return f"\n{table}\n"



"""Download the outputs of a task by ID via CLI."""
import argparse

import inductiva


def download_outputs(args):
    """Download the outputs of a task by ID."""
    task_ids = args.task_id
    filenames = args.filenames
    output_dir = args.output_dir

    if output_dir is not None:
        inductiva.set_output_dir(output_dir)

    if output_dir == "":
        print("`ouput_dir` must not be an empty string.")
        return 1

    if not task_ids:
        print("No ID(s) specified.\n"
              "> Use `inductiva tasks download -h` for help.")
        return 1

    task_ids = set(task_ids)

    for task_id in task_ids:
        try:
            task = inductiva.tasks.Task(task_id)
            task.download_outputs(filenames=filenames)
        except inductiva.client.exceptions.ApiException as exc:
            print(f"Error for task {task_id}:", exc)
    return 0


def register(parser):
    """Register the download task command."""

    subparser = parser.add_parser("download",
                                  help="Download tasks.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "Download the output files produced in the context of the tasks with\n"
        "the given ID(s). All files are downloaded unless the --filenames\n"
        "list is given, in which case only the indicated files for each task\n"
        " are downloaded."
        " The name of the output folder can be configured through the\n"
        "--output_dir option. Files are downloaded to a subdirectory\n"
        "named after the ID of the corresponding task.")

    subparser.add_argument("task_id",
                           type=str,
                           help="ID(s) of the task(s) to download.",
                           nargs="*")
    subparser.add_argument("--filenames",
                           type=str,
                           help="Names of the files to download.",
                           nargs="*")
    subparser.add_argument("--output_dir",
                           type=str,
                           help="Path of where to download the task outputs.")

    subparser.set_defaults(func=download_outputs)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskSubmittedInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "id",
            "is_terminated",
            "status",
        }

        class properties:
            id = schemas.StrSchema

            @staticmethod
            def status() -> typing.Type['TaskStatusCode']:
                return TaskStatusCode

            is_terminated = schemas.BoolSchema

            class position_in_queue(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            TaskPositionInQueue,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'position_in_queue':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class time_to_live_seconds(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_0 = schemas.NumberSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'time_to_live_seconds':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "id": id,
                "status": status,
                "is_terminated": is_terminated,
                "position_in_queue": position_in_queue,
                "time_to_live_seconds": time_to_live_seconds,
            }

    id: MetaOapg.properties.id
    is_terminated: MetaOapg.properties.is_terminated
    status: 'TaskStatusCode'

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["status"]) -> 'TaskStatusCode':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["is_terminated"]
    ) -> MetaOapg.properties.is_terminated:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["position_in_queue"]
    ) -> MetaOapg.properties.position_in_queue:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["time_to_live_seconds"]
    ) -> MetaOapg.properties.time_to_live_seconds:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "status",
        "is_terminated",
        "position_in_queue",
        "time_to_live_seconds",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["status"]) -> 'TaskStatusCode':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["is_terminated"]
    ) -> MetaOapg.properties.is_terminated:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["position_in_queue"]
    ) -> typing.Union[MetaOapg.properties.position_in_queue, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["time_to_live_seconds"]
    ) -> typing.Union[MetaOapg.properties.time_to_live_seconds, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "id",
        "status",
        "is_terminated",
        "position_in_queue",
        "time_to_live_seconds",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        id: typing.Union[
            MetaOapg.properties.id,
            str,
        ],
        is_terminated: typing.Union[
            MetaOapg.properties.is_terminated,
            bool,
        ],
        status: 'TaskStatusCode',
        position_in_queue: typing.Union[MetaOapg.properties.position_in_queue,
                                        dict, frozendict.frozendict, str, date,
                                        datetime, uuid.UUID, int, float,
                                        decimal.Decimal, bool, None, list,
                                        tuple, bytes, io.FileIO,
                                        io.BufferedReader,
                                        schemas.Unset] = schemas.unset,
        time_to_live_seconds: typing.Union[
            MetaOapg.properties.time_to_live_seconds, dict,
            frozendict.frozendict, str, date, datetime, uuid.UUID, int, float,
            decimal.Decimal, bool, None, list, tuple, bytes, io.FileIO,
            io.BufferedReader, schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'TaskSubmittedInfo':
        return super().__new__(
            cls,
            *_args,
            id=id,
            is_terminated=is_terminated,
            status=status,
            position_in_queue=position_in_queue,
            time_to_live_seconds=time_to_live_seconds,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.task_position_in_queue import TaskPositionInQueue
from inductiva.client.model.task_status_code import TaskStatusCode



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class CampaignFullInfo(
    schemas.DictSchema
):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """


    class MetaOapg:
        required = {
            "end_date",
            "capabilities",
            "quotas",
            "credits",
            "begin_date",
            "id",
            "label",
        }
        
        class properties:
            id = schemas.StrSchema
            label = schemas.StrSchema
            credits = schemas.NumberSchema
            begin_date = schemas.DateTimeSchema
            end_date = schemas.DateTimeSchema
            
            
            class capabilities(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    items = schemas.StrSchema
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple[typing.Union[MetaOapg.items, str, ]], typing.List[typing.Union[MetaOapg.items, str, ]]],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'capabilities':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> MetaOapg.items:
                    return super().__getitem__(i)
            
            
            class quotas(
                schemas.ListSchema
            ):
            
            
                class MetaOapg:
                    
                    @staticmethod
                    def items() -> typing.Type['CampaignQuota']:
                        return CampaignQuota
            
                def __new__(
                    cls,
                    _arg: typing.Union[typing.Tuple['CampaignQuota'], typing.List['CampaignQuota']],
                    _configuration: typing.Optional[schemas.Configuration] = None,
                ) -> 'quotas':
                    return super().__new__(
                        cls,
                        _arg,
                        _configuration=_configuration,
                    )
            
                def __getitem__(self, i: int) -> 'CampaignQuota':
                    return super().__getitem__(i)
            __annotations__ = {
                "id": id,
                "label": label,
                "credits": credits,
                "begin_date": begin_date,
                "end_date": end_date,
                "capabilities": capabilities,
                "quotas": quotas,
            }
    
    end_date: MetaOapg.properties.end_date
    capabilities: MetaOapg.properties.capabilities
    quotas: MetaOapg.properties.quotas
    credits: MetaOapg.properties.credits
    begin_date: MetaOapg.properties.begin_date
    id: MetaOapg.properties.id
    label: MetaOapg.properties.label
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["label"]) -> MetaOapg.properties.label: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["credits"]) -> MetaOapg.properties.credits: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["begin_date"]) -> MetaOapg.properties.begin_date: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["end_date"]) -> MetaOapg.properties.end_date: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["capabilities"]) -> MetaOapg.properties.capabilities: ...
    
    @typing.overload
    def __getitem__(self, name: typing_extensions.Literal["quotas"]) -> MetaOapg.properties.quotas: ...
    
    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema: ...
    
    def __getitem__(self, name: typing.Union[typing_extensions.Literal["id", "label", "credits", "begin_date", "end_date", "capabilities", "quotas", ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)
    
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["id"]) -> MetaOapg.properties.id: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["label"]) -> MetaOapg.properties.label: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["credits"]) -> MetaOapg.properties.credits: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["begin_date"]) -> MetaOapg.properties.begin_date: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["end_date"]) -> MetaOapg.properties.end_date: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["capabilities"]) -> MetaOapg.properties.capabilities: ...
    
    @typing.overload
    def get_item_oapg(self, name: typing_extensions.Literal["quotas"]) -> MetaOapg.properties.quotas: ...
    
    @typing.overload
    def get_item_oapg(self, name: str) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]: ...
    
    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal["id", "label", "credits", "begin_date", "end_date", "capabilities", "quotas", ], str]):
        return super().get_item_oapg(name)
    

    def __new__(
        cls,
        *_args: typing.Union[dict, frozendict.frozendict, ],
        end_date: typing.Union[MetaOapg.properties.end_date, str, datetime, ],
        capabilities: typing.Union[MetaOapg.properties.capabilities, list, tuple, ],
        quotas: typing.Union[MetaOapg.properties.quotas, list, tuple, ],
        credits: typing.Union[MetaOapg.properties.credits, decimal.Decimal, int, float, ],
        begin_date: typing.Union[MetaOapg.properties.begin_date, str, datetime, ],
        id: typing.Union[MetaOapg.properties.id, str, ],
        label: typing.Union[MetaOapg.properties.label, str, ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict, frozendict.frozendict, str, date, datetime, uuid.UUID, int, float, decimal.Decimal, None, list, tuple, bytes],
    ) -> 'CampaignFullInfo':
        return super().__new__(
            cls,
            *_args,
            end_date=end_date,
            capabilities=capabilities,
            quotas=quotas,
            credits=credits,
            begin_date=begin_date,
            id=id,
            label=label,
            _configuration=_configuration,
            **kwargs,
        )

from inductiva.client.model.campaign_quota import CampaignQuota



"""CLI for logs."""
import re
import sys
from time import sleep
from typing import Tuple

from inductiva import constants

from .. import utils as cli_utils
from ... import tasks


def _get_task_id_from_mode(mode: str) -> Tuple[bool, str]:
    """Extract the task_id from the mode."""
    match = re.match(r"^(submitted|started)(-(\d+))?$", mode)
    if not match:
        return False, f"Invalid mode format: {mode}"

    offset_match = match.group(3)

    status = match.group(1)
    offset = (int(offset_match) + 1) if offset_match else 1

    # If the status is submitted, send None to the API to get all the tasks.
    # The endpoint already returns the tasks in the submitted order by default
    if status == "submitted":
        status = None

    task_list = tasks.get(last_n=offset, status=status)

    if not task_list:
        return False, f"No '{status}' tasks found."

    return True, task_list[-1].id


def validate_mode_or_task_id(mode: str) -> Tuple[bool, str]:
    """Validate the mode or task_id."""
    if mode.startswith("submitted") or mode.startswith("started"):
        return _get_task_id_from_mode(mode)

    return True, mode


def _check_if_task_is_running(task: tasks.Task, wait: bool) -> Tuple[bool, str]:
    """Check if the task is running.
    
    If the task is not running and the wait flag is set, wait for the task to
    start running.
    
    Args:
        task (tasks.Task): The task object.
        wait (bool): If True, wait for the task to start running.
    """
    info = task.get_info()
    first_time = True

    while wait and not info.is_running and not info.is_terminal:
        if first_time:
            print(f"The task {task.id} whith status '{info.status}' is not "
                  "running yet.\nWaiting for it to start running...")
            first_time = False
        sleep(constants.INDUCTIVA_LOGS_WAIT_SLEEP_TIME)
        info = task.get_info()

    if not info.is_running:
        return False, (
            f"The current status of task {task.id} is '{info.status}'\n"
            "and the simulation logs are not available for streaming.\n"
            "For more information about the task status, use:\n\n"
            f"  inductiva tasks list --id {task.id}\n")

    return True, task.id


def stream_task_logs(args):
    """Consume the stream logs of a certain task."""
    mode = args.mode.lower()
    wait = args.wait
    result, data = validate_mode_or_task_id(mode)
    if not result:
        print(data, file=sys.stderr)
        return 1

    io_stream = None if (
        args.stdout and args.stderr
    ) else "std_out" if args.stdout else "std_err" if args.stderr else None

    no_color = True if io_stream else args.no_color

    task_id = data
    task = tasks.Task(data)

    result, data = _check_if_task_is_running(task, wait=wait)
    if not result:
        print(data, file=sys.stderr)
        return 1

    consumer = tasks.streams.TaskStreamConsumer(task_id,
                                                io_stream=io_stream,
                                                no_color=no_color)
    consumer.run_forever()
    return 0


def register(parser):
    cli_utils.show_help_msg(parser)

    parser.add_argument(
        "mode",
        type=str,
        nargs="?",
        default="SUBMITTED",
        help=
        ("Mode of log retrieval. "
         "Use 'SUBMITTED' for the last submitted task, "
         "'SUBMITTED-1' for the second last submitted task, and so on. "
         "'STARTED' and 'STARTED-n' follow the same pattern for started tasks. "
         "Or, use a specific 'task_id' to retrieve logs for a particular task."
        ))

    parser.add_argument("--stdout",
                        action="store_true",
                        help="Consumes the standard output stream of the task.")
    parser.add_argument("--stderr",
                        action="store_true",
                        help="Consumes the standard error stream of the task.")
    parser.add_argument("--no-color",
                        action="store_true",
                        help="Disables the colorized output.")
    parser.add_argument(
        "--wait",
        "-w",
        action="store_true",
        help=("Wait for the task to start running before consuming the logs.\n"
              "Without this flag, the logs will be consumed immediately\nor "
              "returns an error if the task is not running."))
    # Register function to call when this subcommand is used
    parser.set_defaults(func=stream_task_logs)



"""Available machine types and their number of cores."""
from dataclasses import dataclass
from typing import List, Union
import json

import inductiva
from inductiva.utils import format_utils
from inductiva.client import ApiException
from inductiva.client.apis.tags import compute_api


class ProviderType(format_utils.CaseInsensitiveEnum):
    """Enum to represent the provider of the machine to be launched."""
    GCP = "GCP"
    ICE = "ICE"
    LOCAL = "LOCAL"


@dataclass(frozen=True)
class MachineTypeInfo:
    """Dataclass to represent a machine type.
    This class has all the information related to a machine type.
    """

    threads_per_core: int
    machine_type: str
    provider_id: str
    num_cpus: int
    price: float
    ram_gb: int
    region: str
    spot: bool
    zone: str

    def __getitem__(self, item):
        return getattr(self, item)


def list_available_machines(provider: Union[str, ProviderType]):
    """List all available machines types."""

    resources_available = get_available_machine_types(provider)
    machine_types = []

    for machine in resources_available:
        machine_types.append(machine["machine_type"])

    return tuple(machine_types)


def get_available_machine_types(
        provider: Union[str, ProviderType] = ProviderType.GCP,
        machine_family: str = None) -> List[MachineTypeInfo]:
    """Get all available machine types for a given provider.

    Args:
        provider (str): The provider to list the available machine types.
        machine_family (str): The machine family to filter the specific
            available machine types.
    Returns:
        list (MachineTypeInfo): List of available machine types.
    """

    provider = ProviderType(provider)

    api_client = compute_api.ComputeApi(inductiva.api.get_client())

    query_params = {"provider_id": provider.value}

    if machine_family is not None:
        query_params["machine_family"] = machine_family

    try:
        resp = api_client.list_available_machine_types(query_params).response

        response_body = json.loads(resp.data.decode("utf-8"))
        return [
            MachineTypeInfo(**machine_type) for machine_type in response_body
        ]
    except ApiException as e:
        raise e



from inductiva.client.paths.compute_group.post import ApiForpost
from inductiva.client.paths.compute_group.delete import ApiFordelete


class ComputeGroup(
        ApiForpost,
        ApiFordelete,
):
    pass



"""Configuration for pytest and pytest-markdown."""
import pytest
import inductiva


@pytest.fixture
def simulator():
    fds = inductiva.simulators.FDS()
    return fds


@pytest.fixture
def input_dir():
    input_files_dir = inductiva.utils.download_from_url(
        "https://storage.googleapis.com/inductiva-api-demo-files/"
        "fds-input-example.zip",
        unzip=True)
    return input_files_dir


def pytest_markdown_docs_globals():
    return {"inductiva": inductiva}



"""Inductiva template manager class to manipulate files and render templates."""
import pathlib
import shutil
import os

import jinja2

TEMPLATE_EXTENSION = ".jinja"


class TemplateManager:
    """Template manager for rendering files and directories.

    Template files contains variables that are replaced by
    values provided as keyword arguments to the render methods.
    The underlying template engine is the Jinja2 library.
    """

    @classmethod
    def render_dir(cls,
                   source_dir: str,
                   target_dir: str,
                   overwrite: bool = False,
                   **render_args):
        """
        Renders the entire template directory to the given `target_dir`,
        using the provided rendering args, preserving the directory structure of
        the `source_dir`. Files that are not templates are copied as is.

        When `target_dir` does not exist, it will be created. If any destination
        file inside `target_dir` already exists, a FileExistsError will be
        raised when `overwrite` is False (default). Otherwise, files are
        overwritten.

        Args:
            source_dir (str): Path to the source directory,
                containing template files (and potentially static files to be
                copied without modification).
            target_dir (str): Path to the target directory.
            overwrite (bool): If True, the destination files will be overwritten
                if they already exist. If False (default), a FileExistsError
                will be raised if any destination file already exists.
            render_args: Keyword arguments to render the template files.

        Raises:
            FileExistsError: If the destination file already exists and
                `overwrite` is False.
        """
        if not source_dir:
            raise ValueError("Source directory not provided.")

        source_dir = pathlib.Path(source_dir)
        if not source_dir.is_dir():
            raise ValueError(f"Source directory {source_dir} does not exist.")

        if not target_dir:
            raise ValueError("Target directory not provided.")

        target_dir = pathlib.Path(target_dir)

        loader = jinja2.FileSystemLoader(source_dir)
        environment = jinja2.Environment(loader=loader,
                                         keep_trailing_newline=True,
                                         undefined=jinja2.StrictUndefined)

        template_dir_struct = get_dir_structure(source_dir)

        if not overwrite:
            try:
                check_prerender_dir(template_dir_struct, target_dir)
            except FileExistsError as e:
                msg = f"{e}; set `overwrite=True` to overwrite existing files."
                raise FileExistsError(msg) from e

        for subdir, contents in template_dir_struct.items():
            target_subdir = target_dir / subdir
            target_subdir.mkdir(parents=True, exist_ok=True)
            source_subdir = source_dir / subdir

            # simply copy non-template files to destination
            for file in contents["files"]:
                shutil.copy(source_subdir / file, target_subdir)

            # render template files
            for file in contents["templates"]:
                target_path = target_subdir / strip_extension(file)
                template_name = (pathlib.Path(subdir) / file).as_posix()
                template = environment.get_template(template_name)
                stream = template.stream(**render_args)
                stream.dump(str(target_path))


def is_template(file: str) -> bool:
    """Check if the given file is of template type."""
    return file.endswith(TEMPLATE_EXTENSION)


def strip_extension(file: str) -> str:
    """Strip the template extension if the given file is a template."""
    if not is_template(file):
        return file
    return os.path.splitext(file)[0]


def get_dir_structure(template_dir: str):
    """
    Generate a dictionary with the structure of the template directory.
    The keys are the subdirectories of the template directory, relative to
    the template directory itself. The values are dictionaries with the keys
    "files" and "templates" containing the lists of the files and templates
    in each subdirectory, respectively.
    Example:
    >>> renderer.get_dir_structure("template_dir")
    {
        ".": {
            "files": ["file0.1"],
            "templates": ["template0.1"]
        },
        "subdir1": {
            "files": ["file1.1", "file1.2"],
            "templates": ["template1.1", "template1.2"]
        }
        ...
    }
    """
    return {
        os.path.relpath(root, start=template_dir): {
            "templates": [f for f in files if is_template(f)],
            "files": [f for f in files if not is_template(f)],
        } for root, _, files in os.walk(template_dir, topdown=True)
    }


def check_prerender_dir(source_dir_struct, target_dir: str):
    """Check if the destination filenames exist.

    Check if the destination filenames exist and raise a FileExistsError if
    any of them already exists. For template files, the extension is
    stripped before checking.

    Args:
        source_dir_struct (dict): The structure of the source directory,
            as returned by `get_dir_structure`.
        target_dir (pathlib.Path): The destination directory.
    """
    target_dir = pathlib.Path(target_dir)
    for subdir, contents in source_dir_struct.items():
        dest_subdir = target_dir / subdir
        for file in contents["files"] + contents["templates"]:
            raw_target_name = str(dest_subdir / file)
            target_name = strip_extension(raw_target_name)
            if os.path.exists(target_name):
                raise FileExistsError(f"File {target_name} already exists.")
